-module(lazy).
-compile(export_all).
-compile({no_auto_import,[tl/1]}).
-author('hanak@iit.bme.hu, kapolnai@iit.bme.hu').
-vsn('$LastChangedDate: 2012-12-03 17:04:11 +0100 (h, 03 dec 2012) $$').

% @type lazy:list() = [] | [Head::any()|fun() -> Tail::lazy:list()].
% Félig lusta lista: a fej mindig kiértékelődik, a farok lusta.
% A teljesen lusta lista kicsit komplikáltabb, mert nem tudni, hogy
% üres vagy nem.

% @spec seq(M::integer(), N::integer()) -> lazy:list().
seq(M, N) when M =< N ->
    [M|fun() -> seq(M+1, N) end];
seq(_, _) ->
    [].

% @spec infseq(N::integer()) -> lazy:list().
infseq(N) ->
    [N|fun() -> infseq(N+1) end].

% @spec take(lazy:list(), N::integer()) -> erlang:list().
take(_, 0) ->
    [];
take([H|_], 1) ->  % optimalizáció: hogy ne kelljen a farkát kiértékelni T()
    [H];
take([], _) ->
    [];
take([H|T], N) ->
    [H|take(T(), N-1)].

% Szószátyár take nyomkövetéshez.
take2(_L, 0) ->
    io:format("hivas: take(~w, 0)~n", [_L]), % nyomkoveteshez
    [];
take2([H|_T], 1) ->
    io:format("hivas: take(~w, ~w)~n", [[H|_T], 1]), % nyomkoveteshez
    [H];
take2([], _N) ->
    io:format("hivas: take([], ~w)~n", [_N]), % nyomkoveteshez
    [];
take2([H|T], N) ->
    io:format("hivas: take(~w, ~w)~n", [[H|T], N]), % nyomkoveteshez
    [H|take2(T(), N-1)].


sum(L) ->
    sum(L, 0).

% @spec sum(lazy:list(), number()) -> number().
sum([], X) ->
    X;
sum([H|F], X) ->
    sum(F(), H+X).

% @spec next_cartpow(integer(), erlang:list()) -> erlang:list() | nomore.
next_cartpow(N, [N]) ->
    nomore;
next_cartpow(N, [N|T]) ->
    case next_cartpow(N, T) of
        {ok, C} -> {ok, [0|C]};
        nomore  -> nomore
    end;
next_cartpow(_N, [H|T]) ->
    {ok, [H+1|T]}.

% @spec cartesian_power(integer(), lazy:list()) -> lazy:list().
cartesian_power(N, L) ->
    [L|fun() -> case next_cartpow(N, L) of
                    {ok, C} -> cartesian_power(N, C);
                    nomore  -> []
                end end
    ].

% filter(P::fun(any()) -> bool(), lazy:list()) -> lazy:list().
% kicsit mohó, a lista fejét megkeresi
filter(_, []) ->
    [];
filter(P, [H|T]) ->
    case P(H) of
        true ->
            [H|fun() -> filter(P, T()) end];
        false ->
            filter(P, T()) % megkeresi az első megfelelő elemet
    end.

% @spec pit(integer()) -> lazy:list().
% Pitagoraszi számhármasok N összegig.
pit(N) ->
    filter(fun([A,B,C]) ->
%                   io:format("~p~n",[{A,B,C}]),
                   A+B+C=<N andalso A=<B andalso A*B*C=/=0 andalso A*A+B*B=:=C*C
           end,
           cartesian_power(N,[2,2,3])).

% @spec map(fun(), lazy:list()) -> lazy:list().
map(_, []) ->
    [];
map(F, [H|T]) ->
    [F(H)|fun() -> map(F, T()) end].

% @spec smm() -> lazy:list().
% SEND+MORE=MONEY: 9567+1085=10652
smm() ->
    List2Tuple = fun([S,E,N,D,M,O,R,Y]) -> {S,E,N,D,M,O,R,Y} end,
    map(List2Tuple, 
        filter(fun(L) ->
                       %% io:format("~p~n",[L]),
                       sendmory:check_sum(List2Tuple(L)) andalso sendmory:all_different(L)
               end,
               cartesian_power(9,[0,0,0,7,1,0,8,2])) %% kis csalás, hogy hamar találjon megoldást,
               %% cartesian_power(9,[0,0,0,0,0,0,0,0])) % mert így nagyon sokáig tartana
       ).

% @spec erat() -> lazy:list().
% Eratosztenész szitája.
erat() -> sieve(infseq(2)).

% @spec sift(Prime::integer(), L::lazy:list()) -> L2::lazy:list().
% L2 lista L lista azon elemei, melyek nem oszthatóak Prime-mal.
sift(Prime, L) ->
    filter(fun(N) -> N rem Prime =/= 0 end, L).

% @spec sieve(L1::lazy:list()) -> L2::lazy:list().
% Szitálja az L1 végtelen listát (üres listára hibát ad).
sieve([H|T]) ->
    [H|fun() -> sieve(sift(H, T())) end].

% @spec cons(erlang:list()) -> lazy:list().
cons([]) ->
    [];
cons([H|T]) ->
    [H|fun() -> cons(T) end].

% @spec tl(lazy:list) -> lazy:list.
tl([_|T]) ->
    T().

% @spec fibs(integer(), integer()) -> lazy:list.
fibs(Cur, Next) ->
    [Cur|fun() -> fibs(Next, Cur+Next) end].

% @spec append(lazy:list(), lazy:list()) -> lazy:list().
append([], L2) ->
    L2;
append([H|T], L2) ->
    [H|fun() -> append(T(), L2) end].

% Szószátyár append nyomkövetéshez.
append2([], L2) ->
    io:format("    Hivas: append([], ~w)~n", [(L2)]), % nyomkoveteshez
    L2;
append2([H|T], L2) ->
    io:format("    Hivas: append(~w, ~w)~n", [([H|T]), (L2)]), % nyomkoveteshez
    [H|fun() -> append2(T(), L2) end].

% @spec qsort(lazy:list()) -> lazy:list().
% A lazy:append alkalmazása: ha csak a lista elejére kíváncsi
% a felhasználó, csak a lista elejét rendezi.
qsort([]) ->
    [];
qsort([Pivot|Xs]) ->
    io:format("Hivas: qsort(~w)~n", [take([Pivot|Xs], 100)]), % nyomkoveteshez
    append(qsort(filter(fun(X) -> X < Pivot end, Xs())),
           [Pivot|fun() -> qsort(filter(fun(X) -> X >= Pivot end, Xs())) end]).

% Szószátyár qsort nyomkövetéshez.
qsort2([]) ->
    [];
qsort2([Pivot|Xs]) ->
    io:format("Hivas: qsort(~w)~n", [take([Pivot|Xs], 100)]), % nyomkoveteshez
    Low = fun(X) -> X < Pivot end,
    High = fun(X) -> X >= Pivot end,
    append2(qsort2(filter(Low, Xs())),
           [Pivot|fun() -> qsort2(filter(High, Xs())) end]).


test() ->
    [
     sum(seq(1,100000)),
     sum(cons(lists:seq(1,100000))),
     take([1 | fun() -> [4|fun() -> [9|fun() -> [] end] end] end], 2),
     take([1 | fun() -> [4|fun() -> [9|fun() -> [] end] end] end], 3),
     take(seq(1,3),10),
     take(infseq(1),10),
     tl(tl(infseq(1))),
     take(tl(tl(infseq(1))),1),
     take(fibs(0,1),10),
     take(cartesian_power(3,[1,2]),2),
     take(cartesian_power(3,[1,2]),9),
     take(filter(fun(X) -> X rem 2 =:= 0 end, infseq(3)), 3),
     take(filter(fun(X) -> X rem 2 =:= 7 end, seq(3,6)), 3),
     take(pit(30), 2),
     take(pit(30), 30),
     take(erat(), 10),
     [ take(append(cons("ab"),cons("12")), I)
       || I <- lists:seq(0,4) ],
     take(qsort(cons([5,3,6,8,1,7])), 2),
     take(qsort(cons([5,3,6,8,1,7])), 5),
     take(qsort2(cons([5,3,6,8,1,7])), 2),
     take(qsort2(cons([5,3,6,8,1,7])), 5),
     take(smm(),1),
     smm(),
     vege
    ].

memtest(1) ->
    erlang:garbage_collect(),timer:sleep(1000),statistics(runtime),
    sum(seq(1,10000000)),
    rek:print_mem_usage("seq o sum");
memtest(2) ->
    erlang:garbage_collect(),timer:sleep(1000),statistics(runtime),
    L = lists:seq(1,10000000),
    rek:print_mem_usage("lists:seq"),
    lists:sum(L),
    rek:print_mem_usage("lists:sum").
